package main

import (
	"crypto/rand"
	"crypto/sha256"
	"encoding/base64"
	"flag"
	"fmt"
	"os"
	"strings"

	"github.com/apex/log"

	cliHandler "github.com/apex/log/handlers/cli"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"

	"0xacab.org/leap/menshen/storage"
)

var (
	num     = "number-tokens"
	verbose = "verbose"
	dbFile  = "db-file"
	prefix  = "token-prefix"
	length  = "token-byte-length"
	buckets = "buckets"
)

func main() {
	flag.Int(num, 1, "number of tokens to create")
	flag.Int(length, 16, "number of bytes for token length")

	flag.Bool(verbose, false, "set log verbosity to DEBUG")
	flag.String(dbFile, "./sqlite.db", "location of sqlite database file")
	flag.String(prefix, "solitech_", "auth token prefix")
	flag.String(buckets, "", "comma delimited list of buckets to give token access to")

	pflag.CommandLine.AddGoFlagSet(flag.CommandLine)
	pflag.Parse()
	err := viper.BindPFlags(pflag.CommandLine)
	if err != nil {
		log.WithError(err).Fatal("Failed to BindPFlags")
	}

	verbose := viper.GetBool(verbose)
	if verbose {
		log.SetLevel(log.DebugLevel)
	}
	log.SetHandler(cliHandler.New(os.Stdout))

	dbFile := viper.GetString(dbFile)
	db, err := storage.OpenDatabase(dbFile)
	if err != nil {
		log.WithError(err).Fatal("Failed to OpenDatabase")
	}
	defer db.Close()

	stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
	if err != nil {
		log.WithError(err).Fatal("Failed to prepare db statement")
	}

	num := viper.GetInt(num)
	length := viper.GetInt(length)
	prefix := viper.GetString(prefix)
	buckets := viper.GetString(buckets)

	tokens := []string{}
	for i := 0; i < num; i++ {
		b := make([]byte, length)
		_, err := rand.Read(b)
		if err != nil {
			log.WithError(err).Fatal("Failed rand.Read")
		}
		token := fmt.Sprintf("%v%v", prefix, base64.StdEncoding.EncodeToString(b))
		tokens = append(tokens, token)

		h := sha256.New()
		h.Write([]byte(token))
		authTokenHashBytes := h.Sum(nil)

		authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
		_, err = stmt.Exec(authTokenHashString, buckets)
		if err != nil {
			log.WithError(err).Fatal("Failed to insert token")
		}
	}

	fmt.Println("Tokens:")
	fmt.Printf("%v\n", strings.Join(tokens, "\n"))
}
