package genconfig

// TODO: these functions are hardcoding config for a working openvpn config file, but it's
// a demo for riseup right now.
// We need to be able to adjust by any particular configuration that the service has
// (CA, ciphers etc).

import (
	"bytes"
	"fmt"
	"regexp"
	"strings"
	"text/template"

	"0xacab.org/leap/menshen/pkg/models"
)

type configVars struct {
	CA        string
	Cert      string
	Key       string
	IPAddr    string
	Port      string
	Transport string
}

var openvpnConfigTemplate = `# openvpn config generated by menshen api
client
dev tun
verb 3
tls-client
remote-cert-tls server
tls-version-min 1.2
tls-cipher DHE-RSA-AES128-SHA
cipher AES-256-GCM
pull-filter ignore ifconfig-ipv6
pull-filter ignore route-ipv6
proto {{ .Transport }}
remote {{ .IPAddr }} {{ .Port }} {{ .Transport }}
<ca>
{{ .CA }}
</ca>
<cert>
{{ .Cert }}
</cert>
<key>
{{ .Key }}
</key>`

var (
	rsaBegin  = "-----BEGIN RSA PRIVATE KEY-----"
	rsaEnd    = "-----END RSA PRIVATE KEY-----"
	keyBegin  = "-----BEGIN PRIVATE KEY-----"
	keyEnd    = "-----END PRIVATE KEY-----"
	certBegin = "-----BEGIN CERTIFICATE-----"
	certEnd   = "-----END CERTIFICATE-----"
)

func SerializeConfig(rawCert string, ca string, remote *models.Gateway) (string, error) {
	var key string
	// check if its rsa key or not
	lines := strings.Split(rawCert, "\n")
	if len(lines) == 0 {
		return "", fmt.Errorf("empty cert received")
	}
	if strings.Contains(lines[0], "RSA") {
		key = matchDelimitedString(rawCert, rsaBegin, rsaEnd)
	} else {
		key = matchDelimitedString(rawCert, keyBegin, keyEnd)
	}
	ca = matchDelimitedString(ca, certBegin, certEnd)
	crt := matchDelimitedString(rawCert, certBegin, certEnd)
	vars := configVars{
		CA:        ca,
		Cert:      crt,
		Key:       key,
		IPAddr:    remote.IPAddr,
		Port:      fmt.Sprintf("%d", remote.Port),
		Transport: remote.Transport,
	}
	tmpl, err := template.New("openvpncert").Parse(openvpnConfigTemplate)
	if err != nil {
		return "", err
	}
	var buf bytes.Buffer
	err = tmpl.Execute(&buf, vars)
	if err != nil {
		return "", err
	}
	return buf.String(), nil
}

// matchDelimitedString: extracts the substring between the provided delimiters (left, right) from the main string (ca).
// used for stripping new lines from the extracted substring,
// making it suitable for use cases where newline characters need to be removed.
func matchDelimitedString(str, left, right string) string {
	rx := regexp.MustCompile(`(?s)` + regexp.QuoteMeta(left) + `(.*?)` + regexp.QuoteMeta(right))
	matches := rx.FindAllStringSubmatch(str, -1)
	if len(matches) != 1 {
		return ""
	}
	return left + matches[0][1] + right
}
